/// Copyright (c) 2021 Razeware LLC
///
/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this software and associated documentation files (the "Software"), to deal
/// in the Software without restriction, including without limitation the rights
/// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
/// copies of the Software, and to permit persons to whom the Software is
/// furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// Notwithstanding the foregoing, you may not use, copy, modify, merge, publish,
/// distribute, sublicense, create a derivative work, and/or sell copies of the
/// Software in any work that is designed, intended, or marketed for pedagogical or
/// instructional purposes related to programming, coding, application development,
/// or information technology.  Permission for such use, copying, modification,
/// merger, publication, distribution, sublicensing, creation of derivative works,
/// or sale is expressly withheld.
///
/// This project and source code may use libraries or frameworks that are
/// released under various Open-Source licenses. Use of those libraries and
/// frameworks are governed by their own individual licenses.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
/// THE SOFTWARE.

import simd
import Metal

struct Sphere {
  let buffer: MTLBuffer
  let vertexCount: Int

  // https://stackoverflow.com/a/31326534/359578
  // create vertices using a parametric function

  // the parametric function - this can be any parametric
  // surface function
  // this function is the parametric equation of a sphere
  static func F(_ u: Float, _ v: Float, r: Float = 1) -> simd_float3 {
    [
      cos(u) * sin(v) * r,
      cos(v) * r,
      sin(u) * sin(v) * r
    ]
  }
  
  init(device: MTLDevice, latitude: Int, longitude: Int, radius: Float = 1.0) {
    var vertices: [simd_float3] = []
    let startU: Float = 0
    let startV: Float = 0
    let endU = Float.pi * 2
    let endV = Float.pi
    
    let UResolution: Float = Float(longitude)
    let VResolution: Float = Float(latitude)
    
    let stepU = (endU - startU) / UResolution
    let stepV = (endV - startV) / VResolution
    
    for i in 0..<Int(UResolution) {
      for j in 0..<Int(VResolution) {
        let floatI = Float(i)
        let floatJ = Float(j)
        let u = floatI * stepU + startU
        let v = floatJ * stepV + startV
        
        let un = (floatI + 1 == UResolution) ? endU : (floatI + 1) * stepU + startU
        let vn = (floatJ + 1 == VResolution) ? endV : (floatJ + 1) * stepV + startV
        
        // find the four points of the grid square
        // by evaluating the parametric surface function
        
        let p0 = Self.F(u, v)
        let p1 = Self.F(u, vn)
        let p2 = Self.F(un, v)
        let p3 = Self.F(un, vn)
        
        vertices.append(p0)
        vertices.append(p2)
        vertices.append(p1)
        
        vertices.append(p3)
        vertices.append(p1)
        vertices.append(p2)
      }
    }
    
    guard let buffer = device.makeBuffer(
      bytes: &vertices,
      length: MemoryLayout<simd_float3>.stride * vertices.count,
      options: []) else {
      fatalError("Failed to create buffer")
    }
    self.buffer = buffer
    self.vertexCount = vertices.count
  }
}
